#!/usr/bin/env python3

"""The template engine used to render the data to the destination."""
import sys

from jinja2 import Environment, BaseLoader, ChoiceLoader, FunctionLoader, FileSystemLoader
from typing import Any
from urllib.parse import quote as urlquote

from .config import Context, DEFAULT_TEMPLATE_NAME
from .tree import SourceNode
from ._version import __version__


"""The default template used to render Markdown data."""
DEFAULT_TEMPLATE = """<html lang="{{ lang }}">
    <head>
        {% if site_title is defined %}
            <title>{{ title }} - {{ site_title }}</title>
            <meta property="og:site_name" content="{{ site_title }}"/>
        {% else %}
            <title>{{ title }}</title>
        {% endif %}
        <meta charset="utf-8"/>
        {% if site_base is defined %}
            <base href="{{ site_base }}"/>
        {% endif %}
        <link rel="stylesheet" href="{{ default_stylesheet }}"/>
        {% if description|length > 0 %}
            <meta name="description" content="{{ description }}"/>
            <meta property="og:description" content="{{ description }}"/>
        {% endif %}
        <meta name="generator" content="dragonglass/{{ dragonglass_version }} python/{{ python_version }}"/>
    </head>
    <body>
        <h1>{{ title }}</h1>
        {{ text }}
        {% if backlinks|length > 0 %}
            <div class="generated-header">
                <hr />
                <span>Links to this page</span>
            </div>
            <div class="backlink-items-container">
                {% for item in backlinks %}
                    <div class="backlink-item">
                        <a class="internal-link" href="{{ item.link }}">{{ item.title }}</a>
                    </div>
                {% endfor %}
            </div>
        {% endif %}
        <script src="https://unpkg.com/lucide@latest"></script>
        <script>
            lucide.createIcons();
        </script>
    </body>
</html>
"""


def _create_loader(ctxt: Context) -> BaseLoader:
    """
    Create the loader used to find templates.

    Args:
        ctxt (Context): The context providing the configuration data for dragonglass.

    Returns:
        BaseLoader: A loader to be passed to the template engine.
    """
    choices = []
    if ctxt.source_dir:
        choices.append(FileSystemLoader(ctxt.template_dir))
    choices.append(FunctionLoader(lambda n: DEFAULT_TEMPLATE if n == DEFAULT_TEMPLATE_NAME else None))
    return ChoiceLoader(choices)


def makepath(ctxt: Context, node: SourceNode | None, s: str) -> str:
    """
    Turn a string into a proper path reference, based on relative link settings and prefix.

    Args:
        ctxt (Context): The context to use for the operation.
        node (SourceNode): The current node being processed; if this is ``None``, it's taken from the context.
        s (str): The string to interpret.

    Returns:
        str: The resolved link.
    """
    if not node:
        node = ctxt.current_node
    if ctxt.relative_links:
        in_path = ctxt.source_dir / s
        rel_path = ctxt.source_dir / node.path
        return urlquote(in_path.relative_to(rel_path.parent, walk_up=True).as_posix())
    else:
        return urlquote(ctxt.url_prefix + s)


def create_template_environment(ctxt: Context) -> Environment:
    """
    Create the template environment used to render the data to the destination.

    Args:
        ctxt (Context): The context providing the configuration data for dragonglass.

    Returns:
        Environment: The new template environment.
    """
    tenv = Environment(loader=_create_loader(ctxt))

    # Add global variables easily calculated.
    tenv.globals['dragonglass_version'] = __version__
    tenv.globals["python_version"] = f"{sys.version_info[0]}.{sys.version_info[1]}.{sys.version_info[2]}"

    # Add values derived from configuration.
    tenv.globals["lang"] = ctxt.site_language
    t = ctxt.site_base
    if t:
        tenv.globals["site_base"] = t
    t = ctxt.site_title
    if t:
        tenv.globals["site_title"] = t

    # Add global functions.
    tenv.globals['makepath'] = lambda s: makepath(ctxt, None, s)
    return tenv


def template_vars(node: SourceNode, ctxt: Context) -> dict[str, Any]:
    """
    Create the template variables for the specified node.

    Args:
        node (SourceNode): The node being processed.
        ctxt (Context): The context providing the configuration data for dragonglass.

    Returns:
        dict[str, Any]: The template variables to be used to render the node contents.
    """
    tvars = node.make_vars()

    # Patch description.
    if not tvars["description"] and ctxt.description_title:
        tvars["description"] = node.page_title

    # Compute backlinks for this node.  Don't include nodes that aren't published.
    backlinks = sorted([n for n in node.backlinks if n.publish], key=lambda n: n.page_title)
    tvars['backlinks'] = [{'title': n.page_title,
                           'link': n.link_target(ctxt.url_prefix, node if ctxt.relative_links else None)}
                          for n in backlinks]

    # Add reference to the default stylesheet.
    tvars['default_stylesheet'] = makepath(ctxt, node, ctxt.default_stylesheet)

    # Augment template variables with the extensions.
    ctxt.augment_template_vars(node, tvars)
    return tvars


def print_default() -> None:
    """Prints out the default template."""
    print(DEFAULT_TEMPLATE)
